//! A 256-bit non-cryptographic hash function.
//!
//! This is intended to be used as a fast, high-quality checksum for
//! non-adversarial data identification.  It is not intended to stand
//! up to attacks of any kind.  (It does use the MIX function and
//! constants from Skein v1.3, but is not otherwise related.)
//!
//! This hash does *not* reliably have a full 256 bits worth of power to
//! distinguish different data.  Rather, that number is somewhere
//! between 128 and 192 bits (much closer to the latter).  The 256 bits
//! of output *are*, however, fully diffused.  So you can truncate them
//! to whatever size you like without harm.
//!
//! This implementation should work on platforms of any endianness,
//! but has only been tested on little endian platforms.  Running the
//! unit tests on a big-endian platform can verify.

const BLOCK_SIZE: usize = 256 / 8; // Block size of the hash, in bytes.
const UPDATE_MIX_ROUNDS: usize = 3; // Number of mix rounds after each block of data is added.
const FINISH_MIX_ROUNDS: usize = 6; // Number of mix rounds used to finalize the hash.

/// A hasher.  Consumes bytes and generates a 256-bit hash.
#[derive(Debug, Copy, Clone)]
#[repr(C)]
#[repr(align(32))]
pub struct LedHash256 {
    state: [u64; 4],       // Hash state.
    buf: [u8; BLOCK_SIZE], // Accumulates message data for processing.
    buf_length: usize,     // The number of message bytes currently stored in buf[].
    message_length: u64,   // Accumulates the total message length, in bytes.
}

impl LedHash256 {
    pub fn new() -> LedHash256 {
        LedHash256 {
            state: [
                // Initial Chaining Values from Skein-256-256, v1.3
                0xfc9da860d048b449,
                0x2fca66479fa7d833,
                0xb33bc3896656840f,
                0x6a54e920fde8da69,
            ],
            buf: [0; BLOCK_SIZE],
            buf_length: 0,
            message_length: 0,
        }
    }

    /// Update the hash with new data.
    pub fn update(&mut self, data: &[u8]) {
        self.message_length += data.len() as u64;

        let mut data = data;
        while !data.is_empty() {
            if self.buf_length == BLOCK_SIZE {
                // Process the filled buffer.
                add_data_to_state(&mut self.state, &self.buf);
                mix_state(&mut self.state, UPDATE_MIX_ROUNDS);
                self.buf_length = 0;
            } else if self.buf_length == 0 && data.len() >= BLOCK_SIZE {
                // Process data directly, skipping the buffer.
                add_data_to_state(&mut self.state, data);
                mix_state(&mut self.state, UPDATE_MIX_ROUNDS);
                data = &data[BLOCK_SIZE..];
            } else {
                // Fill the buffer.
                let n = (BLOCK_SIZE - self.buf_length).min(data.len());
                (&mut self.buf[self.buf_length..(self.buf_length + n)]).copy_from_slice(&data[..n]);
                data = &data[n..];
                self.buf_length += n;
            }
        }
    }

    /// Finishes the hash calculations and returns the digest.
    pub fn finish(mut self) -> [u8; BLOCK_SIZE] {
        // Hash the remaining bytes if there are any.
        if self.buf_length > 0 {
            (&mut self.buf[self.buf_length..]).fill(0);
            add_data_to_state(&mut self.state, &self.buf);
            mix_state(&mut self.state, UPDATE_MIX_ROUNDS);
        }

        // Incorporate the message length (in bits) and do the
        // final mixing.
        self.state[0] ^= self.message_length * 8;
        mix_state(&mut self.state, FINISH_MIX_ROUNDS);

        // Get the digest as a byte array and return it.
        let mut digest = [0u8; BLOCK_SIZE];
        digest[0..8].copy_from_slice(&self.state[0].to_le_bytes());
        digest[8..16].copy_from_slice(&self.state[1].to_le_bytes());
        digest[16..24].copy_from_slice(&self.state[2].to_le_bytes());
        digest[24..32].copy_from_slice(&self.state[3].to_le_bytes());
        return digest;
    }
}

/// Adds message data to the hash state.
///
/// The data must be at least 32 bytes long.  Only the first 32 bytes
/// are added.
#[inline(always)]
fn add_data_to_state(state: &mut [u64; 4], data: &[u8]) {
    use std::convert::TryInto;

    // Convert the data to native endian u64's and xor into the
    // hash state.
    assert!(data.len() >= BLOCK_SIZE);
    state[0] ^= u64::from_le_bytes((&data[0..8]).try_into().unwrap());
    state[1] ^= u64::from_le_bytes((&data[8..16]).try_into().unwrap());
    state[2] ^= u64::from_le_bytes((&data[16..24]).try_into().unwrap());
    state[3] ^= u64::from_le_bytes((&data[24..32]).try_into().unwrap());
}

/// Mixes the passed hash state.
///
/// Inspired by Skein 1.3, and using its MIX function and rotation
/// constants.
///
/// Each round actually applies the MIX function twice, with a different
/// word permutation such that each word affects every other word once
/// per round.
///
/// 3 rounds is enough for each bit to have a reasonable chance of
/// affecting most other bits: one input bit flip on average causes about
/// 110 output bit flips.  5 rounds is enough for full diffusion.
#[inline(always)]
fn mix_state(state: &mut [u64; 4], rounds: usize) {
    // Rotation constants from Skein-256-256 v1.3.
    const ROTATIONS: &[[u32; 4]] = &[
        [14, 16, 52, 57],
        [23, 40, 5, 37],
        [25, 33, 46, 12],
        [58, 22, 32, 32],
    ];

    for round in 0..rounds {
        let rot = ROTATIONS[round % ROTATIONS.len()];

        // MIX function.
        state[0] = state[0].wrapping_add(state[2]);
        state[2] = state[2].rotate_left(rot[0]) ^ state[0];
        state[1] = state[1].wrapping_add(state[3]);
        state[3] = state[3].rotate_left(rot[1]) ^ state[1];

        // We change the indices we use below, as if we
        // did a [0 1 2 3] -> [0 1 3 2] permutation.

        // MIX function.
        state[0] = state[0].wrapping_add(state[3]);
        state[3] = state[3].rotate_left(rot[2]) ^ state[0];
        state[1] = state[1].wrapping_add(state[2]);
        state[2] = state[2].rotate_left(rot[3]) ^ state[1];
    }
}

#[cfg(test)]
mod test {
    use super::*;

    fn hash(data: &[u8]) -> [u8; BLOCK_SIZE] {
        let mut h = LedHash256::new();
        h.update(data);
        h.finish()
    }

    fn digest_to_string(digest: [u8; 32]) -> String {
        fn low_bits_to_char(n: u8) -> char {
            match n {
                0 => '0',
                1 => '1',
                2 => '2',
                3 => '3',
                4 => '4',
                5 => '5',
                6 => '6',
                7 => '7',
                8 => '8',
                9 => '9',
                10 => 'a',
                11 => 'b',
                12 => 'c',
                13 => 'd',
                14 => 'e',
                15 => 'f',
                _ => unreachable!(),
            }
        }

        let mut s = String::new();
        for byte in &digest {
            s.push(low_bits_to_char(byte >> 4u8));
            s.push(low_bits_to_char(byte & 0b00001111));
        }
        s
    }

    #[test]
    fn hash_empty() {
        let correct_digest = "fcdfdd47e35abc0d7ebd5c24aaa81b896c07f2cb0f2dc6395fdda6fc8fb12991";
        assert_eq!(digest_to_string(hash(&[])), correct_digest);
    }

    #[test]
    fn hash_zero() {
        let correct_digest = "0f11e90ef9373089f0a337cd1af6c923a2e5d679e92782d3e51da364a34d33e9";
        assert_eq!(digest_to_string(hash(&[0u8])), correct_digest);
    }

    #[test]
    fn hash_one() {
        let correct_digest = "36bf32dc5bcce36f6e1cc268ab40a5d7c1e2ed8dddec59c51a9e79a8a230802f";
        assert_eq!(digest_to_string(hash(&[1u8])), correct_digest);
    }

    #[test]
    fn hash_string_01() {
        let s = "abc";
        let correct_digest = "e06e17ff841570a558f48991172d522b37f86966f19bc45ee7bde2537b212246";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_string_02() {
        let s = "The quick brown fox jumps over the lazy dog.";
        let correct_digest = "e82d9acc9ed0e629115585a253baa4ad607225dcba88cbdb4f1f10979e5b1bfc";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_string_03() {
        let s = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
        let correct_digest = "3e85ef4e523b431b4039bc0b67a8ed80e91be1dc7d650ce6c8a498ba97663cb0";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_string_04() {
        let s = "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.";
        let correct_digest = "793360cfe767f993a3ed4d91238a5042c2b6746c9767cdd77f1b3dbd5f632894";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_multi_part_processing() {
        let test_string1 =
            "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor";
        let test_string2 = " incididunt ut l";
        let test_string3 = "abore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat ";
        let test_string4 = "cup";
        let test_string5 =
            "idatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.";
        let correct_digest = "793360cfe767f993a3ed4d91238a5042c2b6746c9767cdd77f1b3dbd5f632894";

        let mut hasher = LedHash256::new();
        hasher.update(test_string1.as_bytes());
        hasher.update(test_string2.as_bytes());
        hasher.update(test_string3.as_bytes());
        hasher.update(test_string4.as_bytes());
        hasher.update(test_string5.as_bytes());
        let digest = hasher.finish();

        assert_eq!(digest_to_string(digest), correct_digest);
    }

    #[test]
    fn hash_length() {
        // We're testing here to make sure the length of the data properly
        // affects the hash.  Internally in the hash, the last block of data
        // is padded with zeros, so here we're forcing that last block to be
        // all zeros, and only changing the length of input.
        let len_0 = &[];
        let len_1 = &[0u8];
        let len_2 = &[0u8, 0];

        assert_eq!(
            digest_to_string(hash(len_0)),
            "fcdfdd47e35abc0d7ebd5c24aaa81b896c07f2cb0f2dc6395fdda6fc8fb12991",
        );
        assert_eq!(
            digest_to_string(hash(len_1)),
            "0f11e90ef9373089f0a337cd1af6c923a2e5d679e92782d3e51da364a34d33e9",
        );
        assert_eq!(
            digest_to_string(hash(len_2)),
            "70f4c3fdc580f268bd8a81e2163cafa9109ea193f1a062c12ef72996b396043f",
        );
    }
}
